
#ifndef _NEWT_PLUGIN_DETECT_KEYBOARD_H_
#define _NEWT_PLUGIN_DETECT_KEYBOARD_H_

static int
newt_press_key(struct frontend *obj, char *syms, int *codes, int *result);

static int
newt_is_key_there(struct frontend *obj, char *sym, bool plain, bool *result);

#endif /* _NEWT_PLUGIN_DETECT_KEYBOARD_H_ */
